import logging

import nf_core.pipelines.schema


def schema_lint(self):
    """Pipeline schema syntax

    Pipelines should have a ``nextflow_schema.json`` file that describes the different
    pipeline parameters (eg. ``params.something``, ``--something``).

    .. tip:: Reminder: you should generally never need to edit this JSON file by hand.
             The ``nf-core pipelines schema build`` command can create *and edit* the file for you
             to keep it up to date, with a friendly user-interface for customisation.

    The lint test checks the schema for the following:

    * Schema should be a valid JSON file
    * Schema should adhere to `JSONSchema <https://json-schema.org/>`_, Draft 7 or Draft 2020-12.
    * Parameters can be described in two places:

        * As ``properties`` in the top-level schema object
        * As ``properties`` within subschemas listed in a top-level ``definitions``(draft 7) or ``$defs``(draft 2020-12) objects

    * The schema must describe at least one parameter
    * There must be no duplicate parameter IDs across the schema and definition subschema
    * All subschema in ``definitions`` or ``$defs`` must be referenced in the top-level ``allOf`` key
    * The top-level ``allOf`` key must not describe any non-existent definitions
    * Default parameters in the schema must be valid
    * Core top-level schema attributes should exist and be set as follows:

        * ``$schema``: ``https://json-schema.org/draft-07/schema`` or ``https://json-schema.org/draft/2020-12/schema``
        * ``$id``: URL to the raw schema file, eg. ``https://raw.githubusercontent.com/YOURPIPELINE/master/nextflow_schema.json``
        * ``title``: ``YOURPIPELINE pipeline parameters``
        * ``description``: The pipeline config ``manifest.description``
    * That the ``input`` property is defined and has a mimetype. A list of common mimetypes can be found `here <https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types/Common_types>`_.

    For example, an *extremely* minimal schema could look like this (draft 7):

    .. code-block:: json

       {
         "$schema": "https://json-schema.org/draft-07/schema",
         "$id": "https://raw.githubusercontent.com/YOURPIPELINE/master/nextflow_schema.json",
         "title": "YOURPIPELINE pipeline parameters",
         "description": "This pipeline is for testing",
         "properties": {
           "first_param": { "type": "string" }
         },
         "definitions": {
           "my_first_group": {
             "properties": {
               "second_param": { "type": "string" }
             }
           }
         },
         "allOf": [{"$ref": "#/definitions/my_first_group"}]
       }

    Or this (draft 2020-12):

    .. code-block:: json

       {
         "$schema": "https://json-schema.org/draft/2020-12/schema",
         "$id": "https://raw.githubusercontent.com/YOURPIPELINE/master/nextflow_schema.json",
         "title": "YOURPIPELINE pipeline parameters",
         "description": "This pipeline is for testing",
         "properties": {
           "first_param": { "type": "string" }
         },
         "$defs": {
           "my_first_group": {
             "properties": {
               "second_param": { "type": "string" }
             }
           }
         },
         "allOf": [{"$ref": "#/$defs/my_first_group"}]
       }

    .. tip:: You can check your pipeline schema without having to run the entire pipeline lint
             by running ``nf-core pipelines schema lint`` instead of ``nf-core pipelines lint``
    """
    passed = []
    warned = []
    failed = []

    # Only show error messages from schema
    logging.getLogger("nf_core.pipelines.schema").setLevel(logging.ERROR)

    # Lint the schema
    self.schema_obj = nf_core.pipelines.schema.PipelineSchema()
    self.schema_obj.get_schema_path(self.wf_path)

    try:
        self.schema_obj.load_lint_schema()
        passed.append("Schema lint passed")
    except AssertionError as e:
        failed.append(f"Schema lint failed: {e}")

    # Check the title and description - gives warnings instead of fail
    if self.schema_obj.schema is not None:
        try:
            self.schema_obj.validate_schema_title_description()
            passed.append("Schema title + description lint passed")
        except AssertionError as e:
            warned.append(str(e))

    # Check for mimetype in the 'input' parameter, warn if missing
    if self.schema_obj.schema is not None:
        try:
            has_valid_mimetype = self.schema_obj.check_for_input_mimetype()
            if has_valid_mimetype is not None:
                passed.append(f"Input mimetype lint passed: '{has_valid_mimetype}'")
            else:
                warned.append("Input mimetype is missing or empty")
        except LookupError as e:
            warned.append(str(e))

    return {"passed": passed, "warned": warned, "failed": failed}



"""Tests covering the refgenie integration code"""

import os
import shlex
import subprocess
import tempfile
import unittest
from pathlib import Path


class TestRefgenie(unittest.TestCase):
    """Class for refgenie tests"""

    def setUp(self):
        """
        Prepare a refgenie config file
        """
        self.tmp_dir = Path(tempfile.TemporaryDirectory().name)
        self.NXF_HOME = self.tmp_dir / ".nextflow"
        self.NXF_REFGENIE_PATH = self.NXF_HOME / "nf-core" / "refgenie_genomes.config"
        self.REFGENIE = self.tmp_dir / "genomes_config.yaml"
        self.translation_file = self.tmp_dir / "alias_translations.yaml"
        # Set NXF_HOME environment variable
        # avoids adding includeConfig statement to config file outside the current tmpdir
        try:
            self.NXF_HOME_ORIGINAL = os.environ["NXF_HOME"]
        except Exception:
            self.NXF_HOME_ORIGINAL = None
        os.environ["NXF_HOME"] = str(self.NXF_HOME)

        # create NXF_HOME and nf-core directories
        nf_core_dir = self.NXF_HOME / "nf-core"
        nf_core_dir.mkdir(parents=True, exist_ok=True)

        # Initialize a refgenie config
        os.system(f"refgenie init -c {self.REFGENIE}")

        # Add NXF_REFGENIE_PATH to refgenie config
        with open(self.REFGENIE, "a") as fh:
            fh.write(f"nextflow_config: {self.NXF_REFGENIE_PATH}\n")

        # Add an alias translation to YAML file
        with open(self.translation_file, "a") as fh:
            fh.write("ensembl_gtf: gtf\n")

    def tearDown(self) -> None:
        # Reset NXF_HOME environment variable
        if self.NXF_HOME_ORIGINAL is None:
            del os.environ["NXF_HOME"]
        else:
            os.environ["NXF_HOME"] = self.NXF_HOME_ORIGINAL

    def test_update_refgenie_genomes_config(self):
        """Test that listing pipelines works"""
        # Populate the config with a genome
        cmd = f"refgenie pull t7/fasta -c {self.REFGENIE}"
        out = subprocess.check_output(shlex.split(cmd), stderr=subprocess.STDOUT)

        assert "Updated nf-core genomes config" in str(out)

    def test_asset_alias_translation(self):
        """Test that asset aliases are translated correctly"""
        # Populate the config with a genome
        cmd = f"refgenie pull hg38/ensembl_gtf -c {self.REFGENIE}"
        subprocess.check_output(shlex.split(cmd), stderr=subprocess.STDOUT)
        cmd = f"cat {self.NXF_REFGENIE_PATH}"
        out = subprocess.check_output(shlex.split(cmd), stderr=subprocess.STDOUT)
        assert "      gtf                  = " in str(out)
        assert "      ensembl_gtf          = " not in str(out)



"""Tests covering the modules commands"""

import os
import shutil
import tempfile
import unittest
from pathlib import Path

from git.repo import Repo

from .utils import GITLAB_NFTEST_BRANCH, GITLAB_URL


class TestComponents(unittest.TestCase):
    """Class for components tests"""

    def setUp(self):
        """Clone a testing version the nf-core/modules repo"""
        self.tmp_dir = Path(tempfile.mkdtemp())
        self.nfcore_modules = Path(self.tmp_dir, "modules-test")

        Repo.clone_from(GITLAB_URL, self.nfcore_modules, branch=GITLAB_NFTEST_BRANCH)

        # Set $PROFILE environment variable to docker - tests will run with Docker
        if os.environ.get("PROFILE") is None:
            os.environ["PROFILE"] = "docker"

    def tearDown(self):
        """Clean up temporary files and folders"""

        # Clean up temporary files
        if self.tmp_dir.is_dir():
            shutil.rmtree(self.tmp_dir)

    ############################################
    # Test of the individual components commands. #
    ############################################

    from .components.generate_snapshot import (  # type: ignore[misc]
        test_generate_snapshot_module,
        test_generate_snapshot_once,
        test_generate_snapshot_subworkflow,
        test_test_not_found,
        test_unstable_snapshot,
        test_update_snapshot_module,
    )
    from .components.snapshot_test import (  # type: ignore[misc]
        test_components_test_check_inputs,
        test_components_test_no_installed_modules,
        test_components_test_no_name_no_prompts,
    )



from pathlib import Path

from ruamel.yaml import YAML

import nf_core.pipelines.lint

from ..test_lint import TestLint


class TestLintFilesExist(TestLint):
    def setUp(self) -> None:
        super().setUp()
        self.new_pipeline = self._make_pipeline_copy()
        self.lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)

    def test_files_exist_missing_config(self):
        """Lint test: critical files missing FAIL"""

        Path(self.new_pipeline, "CHANGELOG.md").unlink()

        assert self.lint_obj._load()
        self.lint_obj.nf_config["manifest.name"] = "nf-core/testpipeline"

        results = self.lint_obj.files_exist()
        assert "File not found: `CHANGELOG.md`" in results["failed"]

    def test_files_exist_missing_main(self):
        """Check if missing main issues warning"""

        Path(self.new_pipeline, "main.nf").unlink()

        assert self.lint_obj._load()

        results = self.lint_obj.files_exist()
        assert "File not found: `main.nf`" in results["warned"]

    def test_files_exist_deprecated_file(self):
        """Check whether deprecated file issues warning"""

        Path(self.new_pipeline, "parameters.settings.json").touch()

        assert self.lint_obj._load()

        results = self.lint_obj.files_exist()
        assert results["failed"] == ["File must be removed: `parameters.settings.json`"]

    def test_files_exist_pass(self):
        """Lint check should pass if all files are there"""

        assert self.lint_obj._load()

        results = self.lint_obj.files_exist()
        assert results["failed"] == []

    def test_files_exist_pass_conditional_nfschema(self):
        # replace nf-validation with nf-schema in nextflow.config
        with open(Path(self.new_pipeline, "nextflow.config")) as f:
            config = f.read()
        config = config.replace("nf-validation", "nf-schema")
        with open(Path(self.new_pipeline, "nextflow.config"), "w") as f:
            f.write(config)

        assert self.lint_obj._load()
        self.lint_obj.nf_config["manifest.schema"] = "nf-core"
        results = self.lint_obj.files_exist()
        assert results["failed"] == []
        assert results["ignored"] == []

    def test_files_exists_pass_nf_core_yml_config(self):
        """Check if linting passes with a valid nf-core.yml config"""
        valid_yaml = """
        files_exist:
            - .github/CONTRIBUTING.md
            - CITATIONS.md
        """
        yaml = YAML()
        nf_core_yml_path = Path(self.new_pipeline, ".nf-core.yml")
        nf_core_yml = yaml.load(nf_core_yml_path)

        nf_core_yml["lint"] = yaml.load(valid_yaml)
        yaml.dump(nf_core_yml, nf_core_yml_path)

        self.lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        assert self.lint_obj._load()

        results = self.lint_obj.files_exist()
        assert results["failed"] == []
        assert "File is ignored: `.github/CONTRIBUTING.md`" in results["ignored"]
        assert "File is ignored: `CITATIONS.md`" in results["ignored"]

    def test_files_exists_fail_nf_core_yml_config(self):
        """Check if linting fails with a valid nf-core.yml config"""
        valid_yaml = """
        files_exist:
            - CITATIONS.md
        """

        # remove CITATIONS.md
        Path(self.new_pipeline, "CITATIONS.md").unlink()
        assert self.lint_obj._load()
        # test first if linting fails correctly
        results = self.lint_obj.files_exist()
        assert "File not found: `CITATIONS.md`" in results["failed"]

        yaml = YAML()
        nf_core_yml_path = Path(self.new_pipeline, ".nf-core.yml")
        nf_core_yml = yaml.load(nf_core_yml_path)

        nf_core_yml["lint"] = yaml.load(valid_yaml)
        yaml.dump(nf_core_yml, nf_core_yml_path)

        self.lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        assert self.lint_obj._load()

        results = self.lint_obj.files_exist()
        assert results["failed"] == []
        assert "File is ignored: `CITATIONS.md`" in results["ignored"]



#!/usr/bin/env python

from setuptools import find_packages, setup

version = "3.1.1"

with open("README.md") as f:
    readme = f.read()

with open("requirements.txt") as f:
    required = f.read().splitlines()

setup(
    name="nf-core",
    version=version,
    description="Helper tools for use with nf-core Nextflow pipelines.",
    long_description=readme,
    long_description_content_type="text/markdown",
    keywords=[
        "nf-core",
        "nextflow",
        "bioinformatics",
        "workflow",
        "pipeline",
        "biology",
        "sequencing",
        "NGS",
        "next generation sequencing",
    ],
    author="Phil Ewels",
    author_email="phil.ewels@scilifelab.se",
    url="https://github.com/nf-core/tools",
    license="MIT",
    entry_points={
        "console_scripts": ["nf-core=nf_core.__main__:run_nf_core"],
        "refgenie.hooks.post_update": ["nf-core-refgenie=nf_core.pipelines.refgenie:update_config"],
    },
    python_requires=">=3.8, <4",
    install_requires=required,
    packages=find_packages(exclude=("docs")),
    include_package_data=True,
    zip_safe=False,
)



from textwrap import dedent

from textual.app import ComposeResult
from textual.containers import Center
from textual.screen import Screen
from textual.widgets import Button, Footer, Header, Markdown, Static

from nf_core.utils import nfcore_logo

exit_help_text_markdown = """
If you would like to create the GitHub repository later, you can do it manually by following these steps:

1. Create a new GitHub repository
2. Add the remote to your local repository:
    ```bash
    cd <pipeline_directory>
    git remote add origin git@github.com:<username>/<repo_name>.git
    ```
3. Push the code to the remote:
    ```bash
    git push --all origin
    ```
    > 💡 Note the `--all` flag: this is needed to push all branches to the remote.
"""


class GithubExit(Screen):
    """A screen to show a help text when a GitHub repo is NOT created."""

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        yield Markdown(
            dedent(
                """
                # HowTo create a GitHub repository
                """
            )
        )
        yield Static(
            "\n" + "\n".join(nfcore_logo) + "\n",
            id="logo",
        )
        yield Markdown(exit_help_text_markdown)
        yield Center(
            Button("Close", id="close_app", variant="success"),
            classes="cta",
        )



"""
Check whether the content of a module has changed compared to the original repository
"""

import shutil
import tempfile
from pathlib import Path

import nf_core.modules.modules_repo
from nf_core.components.components_differ import ComponentsDiffer


def module_changes(module_lint_object, module):
    """
    Checks whether installed nf-core modules have changed compared to the
    original repository

    Downloads the ``main.nf`` and ``meta.yml`` files for every module
    and compares them to the local copies

    If the module has a commit SHA entry in the ``modules.json``, the file content is
    compared against the files in the remote at this SHA.

    Only runs when linting a pipeline, not the modules repository
    """
    if module.is_patched:
        # If the module is patched, we need to apply
        # the patch in reverse before comparing with the remote
        tempdir_parent = Path(tempfile.mkdtemp())
        tempdir = tempdir_parent / "tmp_module_dir"
        shutil.copytree(module.component_dir, tempdir)
        try:
            new_lines = ComponentsDiffer.try_apply_patch(
                module.component_type,
                module.component_name,
                module.org,
                module.patch_path,
                tempdir,
                reverse=True,
            )
            for file, lines in new_lines.items():
                with open(tempdir / file, "w") as fh:
                    fh.writelines(lines)
        except LookupError:
            # This error is already reported by module_patch, so just return
            return
    else:
        tempdir = module.component_dir
    module.branch = module_lint_object.modules_json.get_component_branch(
        "modules", module.component_name, module.repo_url, module.org
    )
    modules_repo = nf_core.modules.modules_repo.ModulesRepo(remote_url=module.repo_url, branch=module.branch)

    for f, same in modules_repo.component_files_identical(
        module.component_name, tempdir, module.git_sha, "modules"
    ).items():
        if same:
            module.passed.append(
                (
                    "check_local_copy",
                    "Local copy of module up to date",
                    f"{Path(module.component_dir, f)}",
                )
            )
        else:
            module.failed.append(
                (
                    "check_local_copy",
                    "Local copy of module does not match remote",
                    f"{Path(module.component_dir, f)}",
                )
            )



"""
Code for linting subworkflows in the nf-core/subworkflows repository and
in nf-core pipelines

Command:
nf-core subworkflows lint
"""

import logging
import os

import questionary
import rich
import ruamel.yaml

import nf_core.modules.modules_utils
import nf_core.utils
from nf_core.components.lint import ComponentLint, LintExceptionError, LintResult
from nf_core.pipelines.lint_utils import console, run_prettier_on_file

log = logging.getLogger(__name__)

# Import lint functions
from .main_nf import main_nf  # type: ignore[misc]
from .meta_yml import meta_yml  # type: ignore[misc]
from .subworkflow_changes import subworkflow_changes  # type: ignore[misc]
from .subworkflow_tests import subworkflow_tests  # type: ignore[misc]
from .subworkflow_todos import subworkflow_todos  # type: ignore[misc]
from .subworkflow_version import subworkflow_version  # type: ignore[misc]


class SubworkflowLint(ComponentLint):
    """
    An object for linting subworkflows either in a clone of the 'nf-core/modules'
    repository or in any nf-core pipeline directory
    """

    main_nf = main_nf
    meta_yml = meta_yml
    subworkflow_changes = subworkflow_changes
    subworkflow_tests = subworkflow_tests
    subworkflow_todos = subworkflow_todos
    subworkflow_version = subworkflow_version

    def __init__(
        self,
        directory,
        fail_warned=False,
        fix=False,
        remote_url=None,
        branch=None,
        no_pull=False,
        registry=None,
        hide_progress=False,
    ):
        super().__init__(
            component_type="subworkflows",
            directory=directory,
            fail_warned=fail_warned,
            fix=fix,
            remote_url=remote_url,
            branch=branch,
            no_pull=no_pull,
            registry=registry,
            hide_progress=hide_progress,
        )

    def lint(
        self,
        subworkflow=None,
        registry="quay.io",
        key=(),
        all_subworkflows=False,
        print_results=True,
        show_passed=False,
        sort_by="test",
        local=False,
    ):
        """
        Lint all or one specific subworkflow

        First gets a list of all local subworkflows (in subworkflows/local/process) and all subworkflows
        installed from nf-core (in subworkflows/nf-core)

        For all nf-core subworkflows, the correct file structure is assured and important
        file content is verified. If directory subject to linting is a clone of 'nf-core/modules',
        the files necessary for testing the subworkflows are also inspected.

        For all local subworkflows, the '.nf' file is checked for some important flags, and warnings
        are issued if untypical content is found.

        :param subworkflow:     A specific subworkflow to lint
        :param print_results:   Whether to print the linting results
        :param show_passed:     Whether passed tests should be shown as well
        :param hide_progress:   Don't show progress bars

        :returns:               A SubworkflowLint object containing information of
                                the passed, warned and failed tests
        """
        # TODO: consider unifying modules and subworkflows lint() function and add it to the ComponentLint class
        # Prompt for subworkflow or all
        if subworkflow is None and not all_subworkflows:
            questions = [
                {
                    "type": "list",
                    "name": "all_subworkflows",
                    "message": "Lint all subworkflows or a single named subworkflow?",
                    "choices": ["All subworkflows", "Named subworkflow"],
                },
                {
                    "type": "autocomplete",
                    "name": "subworkflow_name",
                    "message": "Subworkflow name:",
                    "when": lambda x: x["all_subworkflows"] == "Named subworkflow",
                    "choices": [m.component_name for m in self.all_remote_components],
                },
            ]
            answers = questionary.unsafe_prompt(questions, style=nf_core.utils.nfcore_question_style)
            all_subworkflows = answers["all_subworkflows"] == "All subworkflows"
            subworkflow = answers.get("subworkflow_name")

        # Only lint the given module
        if subworkflow:
            if all_subworkflows:
                raise LintExceptionError("You cannot specify a tool and request all tools to be linted.")
            local_subworkflows = []
            remote_subworkflows = [s for s in self.all_remote_components if s.component_name == subworkflow]
            if len(remote_subworkflows) == 0:
                raise LintExceptionError(f"Could not find the specified subworkflow: '{subworkflow}'")
        else:
            local_subworkflows = self.all_local_components
            remote_subworkflows = self.all_remote_components

        if self.repo_type == "modules":
            log.info(f"Linting modules repo: [magenta]'{self.directory}'")
        else:
            log.info(f"Linting pipeline: [magenta]'{self.directory}'")
        if subworkflow:
            log.info(f"Linting subworkflow: [magenta]'{subworkflow}'")

        # Filter the tests by the key if one is supplied
        if key:
            self.filter_tests_by_key(key)
            log.info("Only running tests: '{}'".format("', '".join(key)))

        # If it is a pipeline, load the lint config file and the modules.json file
        if self.repo_type == "pipeline":
            self.set_up_pipeline_files()

        # Lint local subworkflows
        if local and len(local_subworkflows) > 0:
            self.lint_subworkflows(local_subworkflows, registry=registry, local=True)

        # Lint nf-core subworkflows
        if len(remote_subworkflows) > 0:
            self.lint_subworkflows(remote_subworkflows, registry=registry, local=False)

        if print_results:
            self._print_results(show_passed=show_passed, sort_by=sort_by)
            self.print_summary()

    def lint_subworkflows(self, subworkflows, registry="quay.io", local=False):
        """
        Lint a list of subworkflows

        Args:
            subworkflows ([NFCoreComponent]): A list of subworkflow objects
            registry (str): The container registry to use. Should be quay.io in most situations.
            local (boolean): Whether the list consist of local or nf-core subworkflows
        """
        # TODO: consider unifying modules and subworkflows lint_subworkflows() function and add it to the ComponentLint class
        progress_bar = rich.progress.Progress(
            "[bold blue]{task.description}",
            rich.progress.BarColumn(bar_width=None),
            "[magenta]{task.completed} of {task.total}[reset] » [bold yellow]{task.fields[test_name]}",
            transient=True,
            console=console,
            disable=self.hide_progress or os.environ.get("HIDE_PROGRESS", None) is not None,
        )
        with progress_bar:
            lint_progress = progress_bar.add_task(
                f"Linting {'local' if local else 'nf-core'} subworkflows",
                total=len(subworkflows),
                test_name=subworkflows[0].component_name,
            )

            for swf in subworkflows:
                progress_bar.update(lint_progress, advance=1, test_name=swf.component_name)
                self.lint_subworkflow(swf, progress_bar, registry=registry, local=local)

    def lint_subworkflow(self, swf, progress_bar, registry, local=False):
        """
        Perform linting on one subworkflow

        If the subworkflow is a local subworkflow we only check the `main.nf` file,
        and issue warnings instead of failures.

        If the subworkflow is a nf-core subworkflow we check for existence of the files
        - main.nf
        - meta.yml
        And verify that their content conform to the nf-core standards.

        If the linting is run for subworkflows in the central nf-core/modules repo
        (repo_type==modules), files that are relevant for subworkflow testing are
        also examined
        """
        # TODO: consider unifying modules and subworkflows lint_subworkflow() function and add it to the ComponentLint class
        # Only check the main script in case of a local subworkflow
        if local:
            self.main_nf(swf)
            self.passed += [LintResult(swf, *s) for s in swf.passed]
            warned = [LintResult(swf, *m) for m in (swf.warned + swf.failed)]
            if not self.fail_warned:
                self.warned += warned
            else:
                self.failed += warned

        # Otherwise run all the lint tests
        else:
            # Update meta.yml file if requested
            if self.fix:
                self.update_meta_yml_file(swf)

            if self.repo_type == "pipeline" and self.modules_json:
                # Set correct sha
                version = self.modules_json.get_subworkflow_version(swf.component_name, swf.repo_url, swf.org)
                swf.git_sha = version

            for test_name in self.lint_tests:
                getattr(self, test_name)(swf)

            self.passed += [LintResult(swf, *s) for s in swf.passed]
            warned = [LintResult(swf, *s) for s in swf.warned]
            if not self.fail_warned:
                self.warned += warned
            else:
                self.failed += warned

            self.failed += [LintResult(swf, *s) for s in swf.failed]

    def update_meta_yml_file(self, swf):
        """
        Update the meta.yml file with the correct inputs and outputs
        """
        yaml = ruamel.yaml.YAML()
        yaml.preserve_quotes = True
        yaml.indent(mapping=2, sequence=2, offset=0)

        # Read meta.yml
        with open(swf.meta_yml) as fh:
            meta_yaml = yaml.load(fh)
        meta_yaml_corrected = meta_yaml.copy()
        # Obtain inputs and outputs from main.nf
        swf.get_inputs_from_main_nf()
        swf.get_outputs_from_main_nf()

        # Compare inputs and add them if missing
        if "input" in meta_yaml:
            # Delete inputs from meta.yml which are not present in main.nf
            meta_yaml_corrected["input"] = [
                input for input in meta_yaml["input"] if list(input.keys())[0] in swf.inputs
            ]
            # Obtain inputs from main.nf missing in meta.yml
            inputs_correct = [
                list(input.keys())[0] for input in meta_yaml_corrected["input"] if list(input.keys())[0] in swf.inputs
            ]
            inputs_missing = [input for input in swf.inputs if input not in inputs_correct]
            # Add missing inputs to meta.yml
            for missing_input in inputs_missing:
                meta_yaml_corrected["input"].append({missing_input: {"description": ""}})

        if "output" in meta_yaml:
            # Delete outputs from meta.yml which are not present in main.nf
            meta_yaml_corrected["output"] = [
                output for output in meta_yaml["output"] if list(output.keys())[0] in swf.outputs
            ]
            # Obtain output from main.nf missing in meta.yml
            outputs_correct = [
                list(output.keys())[0]
                for output in meta_yaml_corrected["output"]
                if list(output.keys())[0] in swf.outputs
            ]
            outputs_missing = [output for output in swf.outputs if output not in outputs_correct]
            # Add missing inputs to meta.yml
            for missing_output in outputs_missing:
                meta_yaml_corrected["output"].append({missing_output: {"description": ""}})

        # Write corrected meta.yml to file
        with open(swf.meta_yml, "w") as fh:
            log.info(f"Updating {swf.meta_yml}")
            yaml.dump(meta_yaml_corrected, fh)
            run_prettier_on_file(fh.name)



from pathlib import Path


def included_configs(self):
    """Check that the pipeline nextflow.config includes the pipeline custom configs.

    If the include line is uncommented, the test passes.
    If the include line is commented, the test fails.
    If the include line is missing, the test warns.

    Can be skipped by adding the following to the .nf-core.yml file:
    lint:
        included_configs: False
    """
    passed = []
    failed = []
    warned = []

    config_file = Path(self.wf_path / "nextflow.config")

    with open(config_file) as fh:
        config = fh.read()
        if (
            f"// includeConfig !System.getenv('NXF_OFFLINE') && params.custom_config_base ? \"${{params.custom_config_base}}/pipeline/{self.pipeline_name}.config\""
            in config
        ):
            failed.append("Pipeline config does not include custom configs. Please uncomment the includeConfig line.")
        elif (
            f"includeConfig !System.getenv('NXF_OFFLINE') && params.custom_config_base ? \"${{params.custom_config_base}}/pipeline/{self.pipeline_name}.config\""
            in config
        ):
            passed.append("Pipeline config includes custom configs.")
        else:
            warned.append("Pipeline config does not include custom configs. Please add the includeConfig line.")

    return {"passed": passed, "failed": failed, "warned": warned}



from pathlib import Path

from ...components.components_differ import ComponentsDiffer
from ...components.nfcore_component import NFCoreComponent


def module_patch(module_lint_obj, module: NFCoreComponent):
    """
    Lint a patch file found in a module

    Checks that the file name is well formed, and that
    the patch can be applied in reverse with the correct result.
    """
    # Check if the module is patched
    if not module.is_patched:
        # Nothing to do
        return

    if not check_patch_valid(module, module.patch_path):
        # Test failed, just exit
        return

    patch_reversible(module_lint_obj, module, module.patch_path)


def check_patch_valid(module, patch_path):
    """
    Checks whether a patch is valid. Looks for lines like
    --- <path>
    +++ <path>
    @@ n,n n,n @@
    and make sure that the come in the right order and that
    the reported paths exists. If the patch file performs
    file creation or deletion we issue a lint warning.

    Args:
        module (NFCoreComponent): The module currently being linted
        patch_path (Path): The absolute path to the patch file.

    Returns:
        (bool): False if any test failed, True otherwise
    """
    with open(patch_path) as fh:
        patch_lines = fh.readlines()

    # Check that the file contains a patch for at least one file
    # and that the file is in the correct directory
    paths_in_patch = []
    passed = True
    it = iter(patch_lines)
    try:
        while True:
            line = next(it)
            if line.startswith("---"):
                frompath = Path(line.split(" ")[1].strip("\n"))
                line = next(it)
                if not line.startswith("+++"):
                    module.failed.append(
                        (
                            "patch_valid",
                            "Patch file invalid. Line starting with '---' should always be followed by line starting with '+++'",
                            patch_path,
                        )
                    )
                    passed = False
                    continue
                topath = Path(line.split(" ")[1].strip("\n"))
                if frompath == Path("/dev/null"):
                    paths_in_patch.append((frompath, ComponentsDiffer.DiffEnum.CREATED))
                elif topath == Path("/dev/null"):
                    paths_in_patch.append((frompath, ComponentsDiffer.DiffEnum.REMOVED))
                elif frompath == topath:
                    paths_in_patch.append((frompath, ComponentsDiffer.DiffEnum.CHANGED))
                else:
                    module.failed.append(
                        (
                            "patch_valid",
                            f"Patch file invaldi. From file '{frompath}' mismatched with to path '{topath}'",
                            patch_path,
                        )
                    )
                    passed = False
                # Check that the next line is hunk
                line = next(it)
                if not line.startswith("@@"):
                    module.failed.append(
                        (
                            "patch_valid",
                            "Patch file invalid. File declarations should be followed by hunk",
                            patch_path,
                        )
                    )
                    passed = False
    except StopIteration:
        pass

    if not passed:
        return False

    if len(paths_in_patch) == 0:
        module.failed.append(("patch_valid", "Patch file invalid. Found no patches", patch_path))
        return False

    # Go through the files and check that they exist
    # Warn about any created or removed files
    passed = True
    for path, diff_status in paths_in_patch:
        if diff_status == ComponentsDiffer.DiffEnum.CHANGED:
            if not Path(module.base_dir, path).exists():
                module.failed.append(
                    (
                        "patch_valid",
                        f"Patch file invalid. Path '{path}' does not exist but is reported in patch file.",
                        patch_path,
                    )
                )
                passed = False
                continue
        elif diff_status == ComponentsDiffer.DiffEnum.CREATED:
            if not Path(module.base_dir, path).exists():
                module.failed.append(
                    (
                        "patch_valid",
                        f"Patch file invalid. Path '{path}' does not exist but is reported in patch file.",
                        patch_path,
                    )
                )
                passed = False
                continue
            module.warned.append(
                ("patch", f"Patch file performs file creation of {path}. This is discouraged."), patch_path
            )
        elif diff_status == ComponentsDiffer.DiffEnum.REMOVED:
            if Path(module.base_dir, path).exists():
                module.failed.append(
                    (
                        "patch_valid",
                        f"Patch file invalid. Path '{path}' is reported as deleted but exists.",
                        patch_path,
                    )
                )
                passed = False
                continue
            module.warned.append(
                ("patch", f"Patch file performs file deletion of {path}. This is discouraged.", patch_path)
            )
        if passed:
            module.passed.append(("patch_valid", "Patch file is valid", patch_path))
        return passed


def patch_reversible(module_lint_object, module, patch_path):
    """
    Try applying a patch in reverse to see if it is up to date

    Args:
        module (NFCoreComponent): The module currently being linted
        patch_path (Path): The absolute path to the patch file.

    Returns:
        (bool): False if any test failed, True otherwise
    """
    try:
        ComponentsDiffer.try_apply_patch(
            module.component_type,
            module.component_name,
            module_lint_object.modules_repo.repo_path,
            patch_path,
            Path(module.component_dir).relative_to(module.base_dir),
            reverse=True,
        )
    except LookupError:
        # Patch failed. Save the patch file by moving to the install dir
        module.failed.append(("patch_reversible", "Patch file is outdated or edited", patch_path))
        return False

    module.passed.append(("patch_reversible", "Patch agrees with module files", patch_path))
    return True



import logging
import sys

import rich

from nf_core.utils import rich_force_colors

log = logging.getLogger(__name__)

stdout = rich.console.Console(force_terminal=rich_force_colors())


def subworkflows_create(ctx, subworkflow, directory, author, force, migrate_pytest):
    """
    Create a new subworkflow from the nf-core template.

    If the specified directory is a pipeline, this function creates a file called
    'subworkflows/local/<subworkflow_name>.nf'

    If the specified directory is a clone of nf-core/modules, it creates or modifies files
    in 'subworkflows/', 'tests/subworkflows' and 'tests/config/pytest_modules.yml'
    """
    from nf_core.subworkflows import SubworkflowCreate

    # Run function
    try:
        subworkflow_create = SubworkflowCreate(directory, subworkflow, author, force, migrate_pytest)
        subworkflow_create.create()
    except UserWarning as e:
        log.critical(e)
        sys.exit(1)
    except LookupError as e:
        log.error(e)
        sys.exit(1)


def subworkflows_test(ctx, subworkflow, directory, no_prompts, update, once, profile, migrate_pytest):
    """
    Run nf-test for a subworkflow.

    Given the name of a subworkflow, runs the nf-test command to test the subworkflow and generate snapshots.
    """
    from nf_core.components.components_test import ComponentsTest

    if migrate_pytest:
        subworkflows_create(ctx, subworkflow, directory, None, False, True)
    try:
        sw_tester = ComponentsTest(
            component_type="subworkflows",
            component_name=subworkflow,
            directory=directory,
            no_prompts=no_prompts,
            update=update,
            once=once,
            remote_url=ctx.obj["modules_repo_url"],
            branch=ctx.obj["modules_repo_branch"],
            verbose=ctx.obj["verbose"],
            profile=profile,
        )
        sw_tester.run()
    except (UserWarning, LookupError) as e:
        log.critical(e)
        sys.exit(1)


def subworkflows_list_remote(ctx, keywords, json):
    """
    List subworkflows in a remote GitHub repo [dim i](e.g [link=https://github.com/nf-core/modules]nf-core/modules[/])[/].
    """
    from nf_core.subworkflows import SubworkflowList

    try:
        subworkflow_list = SubworkflowList(
            ".",
            True,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
        )

        stdout.print(subworkflow_list.list_components(keywords, json))
    except (UserWarning, LookupError) as e:
        log.critical(e)
        sys.exit(1)


def subworkflows_list_local(ctx, keywords, json, directory):  # pylint: disable=redefined-builtin
    """
    List subworkflows installed locally in a pipeline
    """
    from nf_core.subworkflows import SubworkflowList

    try:
        subworkflow_list = SubworkflowList(
            directory,
            False,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
        )
        stdout.print(subworkflow_list.list_components(keywords, json))
    except (UserWarning, LookupError) as e:
        log.error(e)
        sys.exit(1)


def subworkflows_lint(ctx, subworkflow, directory, registry, key, all, fail_warned, local, passed, sort_by, fix):
    """
    Lint one or more subworkflows in a directory.

    Checks DSL2 subworkflow code against nf-core guidelines to ensure
    that all subworkflows follow the same standards.

    Test subworkflows within a pipeline or a clone of the
    nf-core/modules repository.
    """
    from nf_core.components.lint import LintExceptionError
    from nf_core.subworkflows import SubworkflowLint

    try:
        subworkflow_lint = SubworkflowLint(
            directory,
            fail_warned=fail_warned,
            fix=fix,
            registry=ctx.params["registry"],
            remote_url=ctx.obj["modules_repo_url"],
            branch=ctx.obj["modules_repo_branch"],
            no_pull=ctx.obj["modules_repo_no_pull"],
            hide_progress=ctx.obj["hide_progress"],
        )
        subworkflow_lint.lint(
            subworkflow=subworkflow,
            registry=registry,
            key=key,
            all_subworkflows=all,
            print_results=True,
            local=local,
            show_passed=passed,
            sort_by=sort_by,
        )
        if len(subworkflow_lint.failed) > 0:
            sys.exit(1)
    except LintExceptionError as e:
        log.error(e)
        sys.exit(1)
    except (UserWarning, LookupError) as e:
        log.critical(e)
        sys.exit(1)


def subworkflows_info(ctx, subworkflow, directory):
    """
    Show developer usage information about a given subworkflow.

    Parses information from a subworkflow's [i]meta.yml[/] and renders help
    on the command line. A handy equivalent to searching the
    [link=https://nf-co.re/modules]nf-core website[/].

    If run from a pipeline and a local copy of the subworkflow is found, the command
    will print this usage info.
    If not, usage from the remote subworkflows repo will be shown.
    """
    from nf_core.subworkflows import SubworkflowInfo

    try:
        subworkflow_info = SubworkflowInfo(
            directory,
            subworkflow,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
        )
        stdout.print(subworkflow_info.get_component_info())
    except (UserWarning, LookupError) as e:
        log.error(e)
        sys.exit(1)


def subworkflows_install(ctx, subworkflow, directory, prompt, force, sha):
    """
    Install DSL2 subworkflow within a pipeline.

    Fetches and installs subworkflow files from a remote repo e.g. nf-core/modules.
    """
    from nf_core.subworkflows import SubworkflowInstall

    try:
        subworkflow_install = SubworkflowInstall(
            directory,
            force,
            prompt,
            sha,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
        )
        exit_status = subworkflow_install.install(subworkflow)
        if not exit_status:
            sys.exit(1)
    except (UserWarning, LookupError) as e:
        log.error(e)
        sys.exit(1)


def subworkflows_remove(ctx, directory, subworkflow):
    """
    Remove a subworkflow from a pipeline.
    """
    from nf_core.subworkflows import SubworkflowRemove

    try:
        module_remove = SubworkflowRemove(
            directory,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
        )
        module_remove.remove(subworkflow)
    except (UserWarning, LookupError) as e:
        log.critical(e)
        sys.exit(1)


def subworkflows_update(
    ctx,
    subworkflow,
    directory,
    force,
    prompt,
    sha,
    install_all,
    preview,
    save_diff,
    update_deps,
    limit_output,
):
    """
    Update DSL2 subworkflow within a pipeline.

    Fetches and updates subworkflow files from a remote repo e.g. nf-core/modules.
    """
    from nf_core.subworkflows import SubworkflowUpdate

    try:
        subworkflow_install = SubworkflowUpdate(
            directory,
            force,
            prompt,
            sha,
            install_all,
            preview,
            save_diff,
            update_deps,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
            limit_output,
        )
        exit_status = subworkflow_install.update(subworkflow)
        if not exit_status and install_all:
            sys.exit(1)
    except (UserWarning, LookupError) as e:
        log.error(e)
        sys.exit(1)



import shutil

import git
import pytest

import nf_core.pipelines.lint_utils

JSON_WITH_SYNTAX_ERROR = "{'a':1, 1}"
JSON_MALFORMED = "{'a':1}"
JSON_FORMATTED = '{ "a": 1 }\n'

WHICH_PRE_COMMIT = shutil.which("pre-commit")


@pytest.fixture()
def temp_git_repo(tmp_path_factory):
    tmp_git_dir = tmp_path_factory.mktemp("tmp_git_dir")
    repo = git.Repo.init(tmp_git_dir)
    return tmp_git_dir, repo


@pytest.fixture(name="formatted_json")
def git_dir_with_json(temp_git_repo):
    tmp_git_dir, repo = temp_git_repo
    file = tmp_git_dir / "formatted.json"
    with open(file, "w", encoding="utf-8") as f:
        f.write(JSON_FORMATTED)
    repo.git.add(file)
    return file


@pytest.fixture(name="malformed_json")
def git_dir_with_json_malformed(temp_git_repo):
    tmp_git_dir, repo = temp_git_repo
    file = tmp_git_dir / "malformed.json"
    with open(file, "w", encoding="utf-8") as f:
        f.write(JSON_MALFORMED)
    repo.git.add(file)
    return file


@pytest.fixture(name="syntax_error_json")
def git_dir_with_json_syntax_error(temp_git_repo):
    tmp_git_dir, repo = temp_git_repo
    file = tmp_git_dir / "synthax-error.json"
    with open(file, "w", encoding="utf-8") as f:
        f.write(JSON_WITH_SYNTAX_ERROR)
    repo.git.add(file)
    return file


def test_run_prettier_on_formatted_file(formatted_json):
    nf_core.pipelines.lint_utils.run_prettier_on_file(formatted_json)
    assert formatted_json.read_text() == JSON_FORMATTED


def test_run_prettier_on_malformed_file(malformed_json):
    nf_core.pipelines.lint_utils.run_prettier_on_file(malformed_json)
    assert malformed_json.read_text() == JSON_FORMATTED


def test_run_prettier_on_syntax_error_file(syntax_error_json, caplog):
    nf_core.pipelines.lint_utils.run_prettier_on_file(syntax_error_json)
    expected_critical_log = "SyntaxError: Unexpected token (1:10)"
    assert expected_critical_log in caplog.text



import nf_core.pipelines.lint

from ..test_lint import TestLint


class TestLintPluginIncludes(TestLint):
    def setUp(self) -> None:
        super().setUp()
        self.new_pipeline = self._make_pipeline_copy()

    def test_default_values_match(self):
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        result = lint_obj.plugin_includes()
        assert len(result["failed"]) == 0
        assert len(result["warned"]) == 0

    def test_wrong_include(self):
        test_path = self.new_pipeline / "test.nf"
        with open(test_path, "w") as of:
            of.write("include { paramsSummary } from 'plugin/nf-validation'\n")
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        result = lint_obj.plugin_includes()
        assert len(result["failed"]) == 1
        assert len(result["warned"]) == 0



import logging
import os
from pathlib import Path
from typing import List, Optional, Tuple
from urllib.parse import urlparse

from ..components.nfcore_component import NFCoreComponent

log = logging.getLogger(__name__)


class ModuleExceptionError(Exception):
    """Exception raised when there was an error with module commands"""

    pass


def repo_full_name_from_remote(remote_url: str) -> str:
    """
    Extracts the path from the remote URL
    See https://mirrors.edge.kernel.org/pub/software/scm/git/docs/git-clone.html#URLS for the possible URL patterns
    """

    if remote_url.startswith(("https://", "http://", "ftps://", "ftp://", "ssh://")):
        # Parse URL and remove the initial '/'
        path = urlparse(remote_url).path.lstrip("/")
    elif "git@" in remote_url:
        # Extract the part after 'git@' and parse it
        path = urlparse(remote_url.split("git@")[-1]).path
    else:
        path = urlparse(remote_url).path

    # Remove the file extension from the path
    path, _ = os.path.splitext(path)

    return path


def get_installed_modules(directory: Path, repo_type="modules") -> Tuple[List[str], List[NFCoreComponent]]:
    """
    Make a list of all modules installed in this repository

    Returns a tuple of two lists, one for local modules
    and one for nf-core modules. The local modules are represented
    as direct filepaths to the module '.nf' file.
    Nf-core module are returned as file paths to the module directories.
    In case the module contains several tools, one path to each tool directory
    is returned.

    returns (local_modules, nfcore_modules)
    """
    # initialize lists
    local_modules: List[str] = []
    nfcore_modules_names: List[str] = []
    local_modules_dir: Optional[Path] = None
    nfcore_modules_dir = Path(directory, "modules", "nf-core")

    # Get local modules
    if repo_type == "pipeline":
        local_modules_dir = Path(directory, "modules", "local")

        # Filter local modules
        if local_modules_dir.exists():
            local_modules = os.listdir(local_modules_dir)
            local_modules = sorted([x for x in local_modules if x.endswith(".nf")])

    # Get nf-core modules
    if os.path.exists(nfcore_modules_dir):
        for m in sorted([m for m in os.listdir(nfcore_modules_dir) if not m == "lib"]):
            if not os.path.isdir(os.path.join(nfcore_modules_dir, m)):
                raise ModuleExceptionError(
                    f"File found in '{nfcore_modules_dir}': '{m}'! This directory should only contain module directories."
                )
            m_content = os.listdir(os.path.join(nfcore_modules_dir, m))
            # Not a module, but contains sub-modules
            if "main.nf" not in m_content:
                for tool in m_content:
                    nfcore_modules_names.append(os.path.join(m, tool))
            else:
                nfcore_modules_names.append(m)

    # Make full (relative) file paths and create NFCoreComponent objects
    if local_modules_dir:
        local_modules = [os.path.join(local_modules_dir, m) for m in local_modules]

    nfcore_modules = [
        NFCoreComponent(
            m,
            "nf-core/modules",
            Path(nfcore_modules_dir, m),
            repo_type=repo_type,
            base_dir=directory,
            component_type="modules",
        )
        for m in nfcore_modules_names
    ]

    return local_modules, nfcore_modules



import json
import logging
from pathlib import Path

import jsonschema.validators
import yaml

import nf_core.components.components_utils

log = logging.getLogger(__name__)


def meta_yml(subworkflow_lint_object, subworkflow):
    """
    Lint a ``meta.yml`` file

    The lint test checks that the subworkflow has
    a ``meta.yml`` file and that it follows the
    JSON schema defined in the ``subworkflows/yaml-schema.json``
    file in the nf-core/modules repository.

    In addition it checks that the subworkflow name
    and subworkflow input is consistent between the
    ``meta.yml`` and the ``main.nf``.

    Checks that all input and output channels are specified in ``meta.yml``.
    Checks that all included components in ``main.nf`` are specified in ``meta.yml``.

    """
    # Read the meta.yml file
    try:
        with open(subworkflow.meta_yml) as fh:
            meta_yaml = yaml.safe_load(fh)
        subworkflow.passed.append(("meta_yml_exists", "Subworkflow `meta.yml` exists", subworkflow.meta_yml))
    except FileNotFoundError:
        subworkflow.failed.append(("meta_yml_exists", "Subworkflow `meta.yml` does not exist", subworkflow.meta_yml))
        return

    # Confirm that the meta.yml file is valid according to the JSON schema
    valid_meta_yml = True
    try:
        with open(Path(subworkflow_lint_object.modules_repo.local_repo_dir, "subworkflows/yaml-schema.json")) as fh:
            schema = json.load(fh)
        jsonschema.validators.validate(instance=meta_yaml, schema=schema)
        subworkflow.passed.append(("meta_yml_valid", "Subworkflow `meta.yml` is valid", subworkflow.meta_yml))
    except jsonschema.exceptions.ValidationError as e:
        valid_meta_yml = False
        hint = ""
        if len(e.path) > 0:
            hint = f"\nCheck the entry for `{e.path[0]}`."
        if e.message.startswith("None is not of type 'object'") and len(e.path) > 2:
            hint = f"\nCheck that the child entries of {e.path[0]+'.'+e.path[2]} are indented correctly."
        subworkflow.failed.append(
            (
                "meta_yml_valid",
                f"The `meta.yml` of the subworkflow {subworkflow.component_name} is not valid: {e.message}.{hint}",
                subworkflow.meta_yml,
            )
        )
        return

    # Confirm that all input and output channels are specified
    if valid_meta_yml:
        if "input" in meta_yaml:
            meta_input = [list(x.keys())[0] for x in meta_yaml["input"]]
            for input in subworkflow.inputs:
                if input in meta_input:
                    subworkflow.passed.append(("meta_input", f"`{input}` specified", subworkflow.meta_yml))
                else:
                    subworkflow.failed.append(("meta_input", f"`{input}` missing in `meta.yml`", subworkflow.meta_yml))
        else:
            log.debug(f"No inputs specified in subworkflow `main.nf`: {subworkflow.component_name}")

        if "output" in meta_yaml:
            meta_output = [list(x.keys())[0] for x in meta_yaml["output"]]
            for output in subworkflow.outputs:
                if output in meta_output:
                    subworkflow.passed.append(("meta_output", f"`{output}` specified", subworkflow.meta_yml))
                else:
                    subworkflow.failed.append(
                        ("meta_output", f"`{output}` missing in `meta.yml`", subworkflow.meta_yml)
                    )
        else:
            log.debug(f"No outputs specified in subworkflow `main.nf`: {subworkflow.component_name}")

        # confirm that the name matches the process name in main.nf
        if meta_yaml["name"].upper() == subworkflow.workflow_name:
            subworkflow.passed.append(("meta_name", "Correct name specified in `meta.yml`", subworkflow.meta_yml))
        else:
            subworkflow.failed.append(
                (
                    "meta_name",
                    f"Conflicting workflow name between meta.yml (`{meta_yaml['name']}`) and main.nf (`{subworkflow.workflow_name}`)",
                    subworkflow.meta_yml,
                )
            )

        # confirm that all included components in ``main.nf`` are specified in ``meta.yml``
        included_components = nf_core.components.components_utils.get_components_to_install(subworkflow.component_dir)
        included_components = (
            included_components[0] + included_components[1]
        )  # join included modules and included subworkflows in a single list
        if "components" in meta_yaml:
            meta_components = [x for x in meta_yaml["components"]]
            for component in set(included_components):
                if component in meta_components:
                    subworkflow.passed.append(
                        (
                            "meta_include",
                            f"Included module/subworkflow `{component}` specified in `meta.yml`",
                            subworkflow.meta_yml,
                        )
                    )
                else:
                    subworkflow.failed.append(
                        (
                            "meta_include",
                            f"Included module/subworkflow `{component}` missing in `meta.yml`",
                            subworkflow.meta_yml,
                        )
                    )
        if "modules" in meta_yaml:
            subworkflow.failed.append(
                (
                    "meta_modules_deprecated",
                    "Deprecated section 'modules' found in `meta.yml`, use 'components' instead",
                    subworkflow.meta_yml,
                )
            )
        else:
            subworkflow.passed.append(
                (
                    "meta_modules_deprecated",
                    "Deprecated section 'modules' not found in `meta.yml`",
                    subworkflow.meta_yml,
                )
            )



import logging
import os
from pathlib import Path

log = logging.getLogger(__name__)


def modules_structure(self):
    """
    Check that the structure of the modules directory in a pipeline is the correct one:

    .. code-block:: bash

        modules/nf-core/TOOL/SUBTOOL

    Prior to nf-core/tools release 2.6 the directory structure had an additional level of nesting:

    .. code-block:: bash

        modules/nf-core/modules/TOOL/SUBTOOL
    """
    wrong_location_modules = []
    for directory, _, files in os.walk(Path(self.wf_path, "modules")):
        if "main.nf" in files:
            module_path = Path(directory).relative_to(Path(self.wf_path, "modules"))
            parts = module_path.parts
            # Check that there are modules installed directly under the 'modules' directory
            if parts[1] == "modules":
                wrong_location_modules.append(module_path)
    # If there are modules installed in the wrong location
    failed = []
    passed = []
    if len(wrong_location_modules) > 0:
        failed = ["modules directory structure is outdated. Should be 'modules/nf-core/TOOL/SUBTOOL'"]
    else:
        passed = ["modules directory structure is correct 'modules/nf-core/TOOL/SUBTOOL'"]
    return {"passed": passed, "warned": [], "failed": failed, "ignored": []}



import logging

from nf_core.pipelines.lint.pipeline_todos import pipeline_todos

log = logging.getLogger(__name__)


def module_todos(_, module):
    """
    Look for TODO statements in the module files

    The nf-core module template contains a number of comment lines to help developers
    of new modules know where they need to edit files and add content.
    They typically have the following format:

    .. code-block:: groovy

        // TODO nf-core: Make some kind of change to the workflow here

    ..or in markdown:

    .. code-block:: html

        <!-- TODO nf-core: Add some detail to the docs here -->

    This lint test runs through all files in the module and searches for these lines.
    If any are found they will throw a warning.

    .. tip:: Note that many GUI code editors have plugins to list all instances of *TODO*
              in a given project directory. This is a very quick and convenient way to get
              started on your pipeline!

    """

    # Main module directory
    mod_results = pipeline_todos(None, root_dir=module.component_dir)
    for i, warning in enumerate(mod_results["warned"]):
        module.warned.append(("module_todo", warning, mod_results["file_paths"][i]))
    for i, passed in enumerate(mod_results["passed"]):
        module.passed.append(("module_todo", passed, module.component_dir))



import os


def version_consistency(self):
    """Pipeline and container version number consistency.

    .. note:: This test only runs when the ``--release`` flag is set for ``nf-core pipelines lint``,
              or ``$GITHUB_REF`` is equal to ``main``.

    This lint fetches the pipeline version number from three possible locations:

    * The pipeline config, ``manifest.version``
    * The docker container in the pipeline config, ``process.container``

        * Some pipelines may not have this set on a pipeline level. If it is not found, it is ignored.

    * ``$GITHUB_REF``, if it looks like a release tag (``refs/tags/<something>``)

    The test then checks that:

    * The container name has a tag specified (eg. ``nfcore/pipeline:version``)
    * The pipeline version number is numeric (contains only numbers and dots)
    * That the version numbers all match one another
    """
    passed = []
    failed = []

    # Get the version definitions
    # Get version from nextflow.config
    versions = {}
    versions["manifest.version"] = self.nf_config.get("manifest.version", "").strip(" '\"")

    # Get version from the docker tag
    if self.nf_config.get("process.container", "") and ":" not in self.nf_config.get("process.container", ""):
        failed.append(f"Docker slug seems not to have a version tag: {self.nf_config.get('process.container', '')}")

    # Get config container tag (if set; one container per workflow)
    if self.nf_config.get("process.container", ""):
        versions["process.container"] = self.nf_config.get("process.container", "").strip(" '\"").split(":")[-1]

    # Get version from the $GITHUB_REF env var if this is a release
    if (
        os.environ.get("GITHUB_REF", "").startswith("refs/tags/")
        and os.environ.get("GITHUB_REPOSITORY", "") != "nf-core/tools"
    ):
        versions["GITHUB_REF"] = os.path.basename(os.environ["GITHUB_REF"].strip(" '\""))

    # Check if they are all numeric
    for v_type, version in versions.items():
        if not version.replace(".", "").isdigit():
            failed.append(f"{v_type} was not numeric: {version}!")

    # Check if they are consistent
    if len(set(versions.values())) != 1:
        failed.append(
            "The versioning is not consistent between container, release tag " "and config. Found {}".format(
                ", ".join([f"{k} = {v}" for k, v in versions.items()])
            )
        )

    passed.append("Version tags are numeric and consistent between container, release tag and config.")

    return {"passed": passed, "failed": failed}



from pathlib import Path

import yaml

import nf_core.pipelines.bump_version
import nf_core.pipelines.lint

from ..test_lint import TestLint


class TestLintMultiqcConfig(TestLint):
    def setUp(self) -> None:
        super().setUp()
        self.new_pipeline = self._make_pipeline_copy()
        self.multiqc_config_yml = Path(self.new_pipeline, "assets", "multiqc_config.yml")

    def test_multiqc_config_exists(self):
        """Test that linting fails if the multiqc_config.yml file is missing"""
        # Delete the file
        self.multiqc_config_yml.unlink()
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj._load()
        result = lint_obj.multiqc_config()
        assert result["failed"] == ["`assets/multiqc_config.yml` not found."]

    def test_multiqc_config_ignore(self):
        """Test that linting succeeds if the multiqc_config.yml file is missing but ignored"""
        # Delete the file
        self.multiqc_config_yml.unlink()
        with open(Path(self.new_pipeline, ".nf-core.yml")) as f:
            content = yaml.safe_load(f)
            old_content = content.copy()
            content["lint"] = {"multiqc_config": False}
        with open(Path(self.new_pipeline, ".nf-core.yml"), "w") as f:
            yaml.dump(content, f)

        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj._load()
        result = lint_obj.multiqc_config()
        assert result["ignored"] == ["`assets/multiqc_config.yml` not found, but it is ignored."]

        # cleanup
        with open(Path(self.new_pipeline, ".nf-core.yml"), "w") as f:
            yaml.dump(old_content, f)

    def test_multiqc_config_missing_report_section_order(self):
        """Test that linting fails if the multiqc_config.yml file is missing the report_section_order"""
        with open(self.multiqc_config_yml) as fh:
            mqc_yml = yaml.safe_load(fh)
        mqc_yml_tmp = mqc_yml.copy()
        mqc_yml.pop("report_section_order")
        with open(self.multiqc_config_yml, "w") as fh:
            yaml.safe_dump(mqc_yml, fh)
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj._load()
        result = lint_obj.multiqc_config()
        # Reset the file
        with open(self.multiqc_config_yml, "w") as fh:
            yaml.safe_dump(mqc_yml_tmp, fh)
        assert result["failed"] == ["`assets/multiqc_config.yml` does not contain `report_section_order`"]

    def test_multiqc_incorrect_export_plots(self):
        """Test that linting fails if the multiqc_config.yml file has an incorrect value for export_plots"""
        with open(self.multiqc_config_yml) as fh:
            mqc_yml = yaml.safe_load(fh)
        mqc_yml_tmp = mqc_yml.copy()
        mqc_yml["export_plots"] = False
        with open(self.multiqc_config_yml, "w") as fh:
            yaml.safe_dump(mqc_yml, fh)
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj._load()
        result = lint_obj.multiqc_config()
        # Reset the file
        with open(self.multiqc_config_yml, "w") as fh:
            yaml.safe_dump(mqc_yml_tmp, fh)
        assert result["failed"] == ["`assets/multiqc_config.yml` does not contain 'export_plots: true'."]

    def test_multiqc_config_report_comment_fail(self):
        """Test that linting fails if the multiqc_config.yml file has an incorrect report_comment"""
        with open(self.multiqc_config_yml) as fh:
            mqc_yml = yaml.safe_load(fh)
        mqc_yml_tmp = mqc_yml.copy()
        mqc_yml["report_comment"] = "This is a test"
        with open(self.multiqc_config_yml, "w") as fh:
            yaml.safe_dump(mqc_yml, fh)
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj._load()
        result = lint_obj.multiqc_config()
        # Reset the file
        with open(self.multiqc_config_yml, "w") as fh:
            yaml.safe_dump(mqc_yml_tmp, fh)
        assert len(result["failed"]) == 1
        assert result["failed"][0].startswith(
            "`assets/multiqc_config.yml` does not contain a matching 'report_comment'."
        )

    def test_multiqc_config_report_comment_release_fail(self):
        """Test that linting fails if the multiqc_config.yml file has an incorrect report_comment for a release version"""
        with open(self.multiqc_config_yml) as fh:
            mqc_yml = yaml.safe_load(fh)
        mqc_yml_tmp = mqc_yml.copy()
        with open(self.multiqc_config_yml, "w") as fh:
            yaml.safe_dump(mqc_yml, fh)
        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj._load()
        # bump version
        lint_obj.nf_config["manifest.version"] = "1.0"
        result = lint_obj.multiqc_config()
        # Reset the file
        with open(self.multiqc_config_yml, "w") as fh:
            yaml.safe_dump(mqc_yml_tmp, fh)
        assert len(result["failed"]) == 1
        assert result["failed"][0].startswith(
            "`assets/multiqc_config.yml` does not contain a matching 'report_comment'."
        )

    def test_multiqc_config_report_comment_release_succeed(self):
        """Test that linting fails if the multiqc_config.yml file has a correct report_comment for a release version"""

        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj._load()
        # bump version using the bump_version function
        nf_core.pipelines.bump_version.bump_pipeline_version(lint_obj, "1.0")
        # lint again
        lint_obj._load()
        result = lint_obj.multiqc_config()
        assert "`assets/multiqc_config.yml` contains a matching 'report_comment'." in result["passed"]



import difflib
import enum
import json
import logging
import os
from pathlib import Path
from typing import Dict, List, Union

from rich import box
from rich.console import Console, Group, RenderableType
from rich.panel import Panel
from rich.syntax import Syntax

import nf_core.utils

log = logging.getLogger(__name__)


class ComponentsDiffer:
    """
    Static class that provides functionality for computing diffs between
    different instances of a module or subworkflow
    """

    class DiffEnum(enum.Enum):
        """
        Enumeration for keeping track of
        the diff status of a pair of files
        """

        UNCHANGED = enum.auto()
        CHANGED = enum.auto()
        CREATED = enum.auto()
        REMOVED = enum.auto()

    @staticmethod
    def get_component_diffs(from_dir, to_dir, for_git=True, dsp_from_dir=None, dsp_to_dir=None):
        """
        Compute the diff between the current component version
        and the new version.

        Args:
            from_dir (strOrPath): The folder containing the old component files
            to_dir (strOrPath): The folder containing the new component files
            path_in_diff (strOrPath): The directory displayed containing the component
                                      file in the diff. Added so that temporary dirs
                                      are not shown
            for_git (bool): indicates whether the diff file is to be
                            compatible with `git apply`. If true it
                            adds a/ and b/ prefixes to the file paths
            dsp_from_dir (str | Path): The from directory to display in the diff
            dsp_to_dir (str | Path): The to directory to display in the diff

        Returns:
            dict[str, (ComponentsDiffer.DiffEnum, str)]: A dictionary containing
            the diff type and the diff string (empty if no diff)
        """
        if for_git:
            dsp_from_dir = Path("a", dsp_from_dir)
            dsp_to_dir = Path("b", dsp_to_dir)

        diffs = {}
        # Get all unique filenames in the two folders.
        # `dict.fromkeys()` is used instead of `set()` to preserve order
        files = dict.fromkeys(
            [Path(dirpath, file).relative_to(to_dir) for dirpath, _, files in os.walk(to_dir) for file in files]
        )
        files.update(
            dict.fromkeys(
                [Path(dirpath, file).relative_to(from_dir) for dirpath, _, files in os.walk(from_dir) for file in files]
            )
        )
        files = list(files)

        # Loop through all the component files and compute their diffs if needed
        for file in files:
            temp_path = Path(to_dir, file)
            curr_path = Path(from_dir, file)
            if temp_path.exists() and curr_path.exists() and temp_path.is_file():
                with open(temp_path) as fh:
                    new_lines = fh.readlines()
                with open(curr_path) as fh:
                    old_lines = fh.readlines()

                if new_lines == old_lines:
                    # The files are identical
                    diffs[file] = (ComponentsDiffer.DiffEnum.UNCHANGED, ())
                else:
                    # Compute the diff
                    diff = difflib.unified_diff(
                        old_lines,
                        new_lines,
                        fromfile=str(Path(dsp_from_dir, file)),
                        tofile=str(Path(dsp_to_dir, file)),
                    )
                    diffs[file] = (ComponentsDiffer.DiffEnum.CHANGED, diff)

            elif temp_path.exists():
                with open(temp_path) as fh:
                    new_lines = fh.readlines()
                # The file was created
                # Show file against /dev/null
                diff = difflib.unified_diff(
                    [],
                    new_lines,
                    fromfile=str(Path("/dev", "null")),
                    tofile=str(Path(dsp_to_dir, file)),
                )
                diffs[file] = (ComponentsDiffer.DiffEnum.CREATED, diff)

            elif curr_path.exists():
                # The file was removed
                # Show file against /dev/null
                with open(curr_path) as fh:
                    old_lines = fh.readlines()
                diff = difflib.unified_diff(
                    old_lines,
                    [],
                    fromfile=str(Path(dsp_from_dir, file)),
                    tofile=str(Path("/dev", "null")),
                )
                diffs[file] = (ComponentsDiffer.DiffEnum.REMOVED, diff)

        return diffs

    @staticmethod
    def write_diff_file(
        diff_path,
        component,
        repo_path,
        from_dir,
        to_dir,
        current_version=None,
        new_version=None,
        file_action="a",
        for_git=True,
        dsp_from_dir=None,
        dsp_to_dir=None,
        limit_output=False,
    ):
        """
        Writes the diffs of a component to the diff file.

        Args:
            diff_path (str | Path): The path to the file that should be appended
            component (str): The component name
            repo_path (str): The name of the repo where the component resides
            from_dir (str | Path): The directory containing the old component files
            to_dir (str | Path): The directory containing the new component files
            diffs (dict[str, (ComponentsDiffer.DiffEnum, str)]): A dictionary containing
                                                              the type of change and
                                                              the diff (if any)
            current_version (str): The installed version of the component
            new_version (str): The version of the component the diff is computed against
            for_git (bool): indicates whether the diff file is to be
                            compatible with `git apply`. If true it
                            adds a/ and b/ prefixes to the file paths
            dsp_from_dir (str | Path): The 'from' directory displayed in the diff
            dsp_to_dir (str | Path): The 'to' directory displayed in the diff
            limit_output (bool): If true, don't write the diff for files other than main.nf
        """
        if dsp_from_dir is None:
            dsp_from_dir = from_dir
        if dsp_to_dir is None:
            dsp_to_dir = to_dir

        diffs = ComponentsDiffer.get_component_diffs(from_dir, to_dir, for_git, dsp_from_dir, dsp_to_dir)
        if all(diff_status == ComponentsDiffer.DiffEnum.UNCHANGED for _, (diff_status, _) in diffs.items()):
            raise UserWarning("Component is unchanged")
        log.debug(f"Writing diff of '{component}' to '{diff_path}'")
        with open(diff_path, file_action) as fh:
            if current_version is not None and new_version is not None:
                fh.write(
                    f"Changes in component '{Path(repo_path, component)}' between"
                    f" ({current_version}) and"
                    f" ({new_version})\n"
                )
            else:
                fh.write(f"Changes in component '{Path(repo_path, component)}'\n")

            for file, (diff_status, diff) in diffs.items():
                if diff_status == ComponentsDiffer.DiffEnum.UNCHANGED:
                    # The files are identical
                    fh.write(f"'{Path(dsp_from_dir, file)}' is unchanged\n")
                elif diff_status == ComponentsDiffer.DiffEnum.CREATED:
                    # The file was created between the commits
                    fh.write(f"'{Path(dsp_from_dir, file)}' was created\n")
                elif diff_status == ComponentsDiffer.DiffEnum.REMOVED:
                    # The file was removed between the commits
                    fh.write(f"'{Path(dsp_from_dir, file)}' was removed\n")
                elif limit_output and not file.suffix == ".nf":
                    # Skip printing the diff for files other than main.nf
                    fh.write(f"Changes in '{Path(component, file)}' but not shown\n")
                else:
                    # The file has changed write the diff lines to the file
                    fh.write(f"Changes in '{Path(component, file)}':\n")
                    for line in diff:
                        fh.write(line)
                    fh.write("\n")

            fh.write("*" * 60 + "\n")

    @staticmethod
    def append_modules_json_diff(diff_path, old_modules_json, new_modules_json, modules_json_path, for_git=True):
        """
        Compare the new modules.json and builds a diff

        Args:
            diff_fn (str): The diff file to be appended
            old_modules_json (nested dict): The old modules.json
            new_modules_json (nested dict): The new modules.json
            modules_json_path (str): The path to the modules.json
            for_git (bool): indicates whether the diff file is to be
                            compatible with `git apply`. If true it
                            adds a/ and b/ prefixes to the file paths
        """
        fromfile = modules_json_path
        tofile = modules_json_path
        if for_git:
            fromfile = Path("a", fromfile)
            tofile = Path("b", tofile)

        modules_json_diff = difflib.unified_diff(
            json.dumps(old_modules_json, indent=4).splitlines(keepends=True),
            json.dumps(new_modules_json, indent=4).splitlines(keepends=True),
            fromfile=str(fromfile),
            tofile=str(tofile),
        )

        # Save diff for modules.json to file
        with open(diff_path, "a") as fh:
            fh.write("Changes in './modules.json'\n")
            for line in modules_json_diff:
                fh.write(line)
            fh.write("*" * 60 + "\n")

    @staticmethod
    def print_diff(
        component,
        repo_path,
        from_dir,
        to_dir,
        current_version=None,
        new_version=None,
        dsp_from_dir=None,
        dsp_to_dir=None,
        limit_output=False,
    ):
        """
        Prints the diffs between two component versions to the terminal

        Args:
            component (str): The component name
            repo_path (str): The name of the repo where the component resides
            from_dir (str | Path): The directory containing the old component files
            to_dir (str | Path): The directory containing the new component files
            current_version (str): The installed version of the component
            new_version (str): The version of the component the diff is computed against
            dsp_from_dir (str | Path): The 'from' directory displayed in the diff
            dsp_to_dir (str | Path): The 'to' directory displayed in the diff
            limit_output (bool): If true, don't print the diff for files other than main.nf
        """
        if dsp_from_dir is None:
            dsp_from_dir = from_dir
        if dsp_to_dir is None:
            dsp_to_dir = to_dir

        diffs = ComponentsDiffer.get_component_diffs(
            from_dir, to_dir, for_git=False, dsp_from_dir=dsp_from_dir, dsp_to_dir=dsp_to_dir
        )
        console = Console(force_terminal=nf_core.utils.rich_force_colors())
        if current_version is not None and new_version is not None:
            log.info(
                f"Changes in component '{Path(repo_path, component)}' between"
                f" ({current_version}) and"
                f" ({new_version})"
            )
        else:
            log.info(f"Changes in component '{Path(repo_path, component)}'")

        panel_group: list[RenderableType] = []
        for file, (diff_status, diff) in diffs.items():
            if diff_status == ComponentsDiffer.DiffEnum.UNCHANGED:
                # The files are identical
                log.info(f"'{Path(dsp_from_dir, file)}' is unchanged")
            elif diff_status == ComponentsDiffer.DiffEnum.CREATED:
                # The file was created between the commits
                log.info(f"'{Path(dsp_from_dir, file)}' was created")
            elif diff_status == ComponentsDiffer.DiffEnum.REMOVED:
                # The file was removed between the commits
                log.info(f"'{Path(dsp_from_dir, file)}' was removed")
            elif limit_output and not file.suffix == ".nf":
                # Skip printing the diff for files other than main.nf
                log.info(f"Changes in '{Path(component, file)}' but not shown")
            else:
                # The file has changed
                log.info(f"Changes in '{Path(component, file)}':")
                # Pretty print the diff using the pygments diff lexer
                syntax = Syntax("".join(diff), "diff", theme="ansi_dark", line_numbers=True)
                panel_group.append(Panel(syntax, title=str(file), title_align="left", padding=0))
            console.print(
                Panel(
                    Group(*panel_group),
                    title=f"[white]{str(component)}[/white]",
                    title_align="left",
                    padding=0,
                    border_style="blue",
                    box=box.HEAVY,
                )
            )

    @staticmethod
    def per_file_patch(patch_fn: Union[str, Path]) -> Dict[str, List[str]]:
        """
        Splits a patch file for several files into one patch per file.

        Args:
            patch_fn (str | Path): The path to the patch file

        Returns:
            dict[str, str]: A dictionary indexed by the filenames with the
                            file patches as values
        """
        with open(patch_fn) as fh:
            lines = fh.readlines()

        patches = {}
        i = 0
        patch_lines: List[str] = []
        key = "preamble"
        while i < len(lines):
            line = lines[i]
            if line.startswith("---"):
                # New file found: add the old lines to the dictionary
                # and determine the new filename
                patches[key] = patch_lines
                _, frompath = line.split(" ")
                frompath = frompath.strip()
                patch_lines = [line]
                i += 1
                line = lines[i]
                if not line.startswith("+++ "):
                    raise LookupError("Missing to-file in patch file")
                _, topath = line.split(" ")
                topath = topath.strip()
                patch_lines.append(line)

                if frompath == topath:
                    key = frompath
                elif frompath == "/dev/null":
                    key = topath
                else:
                    key = frompath
            else:
                patch_lines.append(line)
            i += 1
        patches[key] = patch_lines

        # Remove the 'preamble' key (entry contains no useful information)
        patches.pop("preamble")
        return patches

    @staticmethod
    def get_new_and_old_lines(patch):
        """
        Parse a patch for a file, and return the contents
        of the modified parts for both the old and new versions

        Args:
            patch (str): The patch in unified diff format

        Returns:
            ([[str]], [[str]]): Lists of old and new lines for each hunk
                                (modified part the file)
        """
        old_lines = []
        new_lines = []
        old_partial = []
        new_partial = []
        # First two lines indicate the file names, the third line is the first hunk
        for line in patch[3:]:
            if line.startswith("@"):
                old_lines.append(old_partial)
                new_lines.append(new_partial)
                old_partial = []
                new_partial = []
            elif line.startswith(" "):
                # Line belongs to both files
                line = line[1:]
                old_partial.append(line)
                new_partial.append(line)
            elif line.startswith("+"):
                # Line only belongs to the new file
                line = line[1:]
                new_partial.append(line)
            elif line.startswith("-"):
                # Line only belongs to the old file
                line = line[1:]
                old_partial.append(line)
        old_lines.append(old_partial)
        new_lines.append(new_partial)
        return old_lines, new_lines

    @staticmethod
    def try_apply_single_patch(file_lines, patch, reverse=False):
        """
        Tries to apply a patch to a modified file. Since the line numbers in
        the patch do not agree if the file is modified, the old and new
        lines in the patch are reconstructed and then we look for the old lines
        in the modified file. If all hunk in the patch are found in the new file
        it is updated with the new lines from the patch file.

        Args:
            file_lines ([str]): The lines of the file to be patched
            patch (str | Path): (Outdated) patch for the file
            reverse (bool): Apply the patch in reverse

        Returns:
            [str]: The patched lines of the file

        Raises:
            LookupError: If it fails to find the old lines from the patch in
                         the file.
        """
        org_lines, patch_lines = ComponentsDiffer.get_new_and_old_lines(patch)
        if reverse:
            patch_lines, org_lines = org_lines, patch_lines

        # The patches are sorted by their order of occurrence in the original
        # file. Loop through the new file and try to find the new indices of
        # these lines. We know they are non overlapping, and thus only need to
        # look at the file once
        p = len(org_lines)
        patch_indices = [None] * p
        i = 0
        j = 0
        n = len(file_lines)
        while i < n and j < p:
            m = len(org_lines[j])
            while i < n:
                if org_lines[j] == file_lines[i : i + m]:
                    patch_indices[j] = (i, i + m)
                    j += 1
                    break
                else:
                    i += 1

        if j != len(org_lines):
            # We did not find all diffs before we ran out of file.
            raise LookupError("Failed to find lines where patch should be applied")

        # Apply the patch to new lines by substituting
        # the original lines with the patch lines
        patched_new_lines = file_lines[: patch_indices[0][0]]
        for i in range(len(patch_indices) - 1):
            # Add the patch lines
            patched_new_lines.extend(patch_lines[i])
            # Fill the spaces between the patches
            patched_new_lines.extend(file_lines[patch_indices[i][1] : patch_indices[i + 1][0]])

        # Add the remaining part of the new file
        patched_new_lines.extend(patch_lines[-1])
        patched_new_lines.extend(file_lines[patch_indices[-1][1] :])

        return patched_new_lines

    @staticmethod
    def try_apply_patch(
        component_type: str,
        component: str,
        repo_path: Union[str, Path],
        patch_path: Union[str, Path],
        component_dir: Path,
        reverse: bool = False,
    ) -> Dict[str, List[str]]:
        """
        Try applying a full patch file to a module or subworkflow

        Args:
            component_type (str): The type of component (modules or subworkflows)
            component (str): Name of the module or subworkflow
            repo_path (str): Name of the repository where the component resides
            patch_path (str): The absolute path to the patch file to be applied
            component_dir (Path): The directory containing the component
            reverse (bool): Apply the patch in reverse

        Returns:
            dict[str, str]: A dictionary with file paths (relative to the pipeline dir)
                            as keys and the patched file contents as values

        Raises:
            LookupError: If the patch application fails in a file
        """
        component_relpath = Path(component_type, repo_path, component)
        patches = ComponentsDiffer.per_file_patch(patch_path)
        new_files = {}
        for file, patch in patches.items():
            log.debug(f"Applying patch to {file}")
            fn = Path(file).relative_to(component_relpath)
            file_path = component_dir / fn
            try:
                with open(file_path) as fh:
                    file_lines = fh.readlines()
            except FileNotFoundError:
                # The file was added with the patch
                file_lines = [""]
            patched_new_lines = ComponentsDiffer.try_apply_single_patch(file_lines, patch, reverse=reverse)
            new_files[str(fn)] = patched_new_lines
        return new_files



import logging
import os
from pathlib import Path
from textwrap import dedent

import git
import yaml
from github import Github, GithubException, UnknownObjectException
from textual import on, work
from textual.app import ComposeResult
from textual.containers import Center, Horizontal, Vertical
from textual.message import Message
from textual.screen import Screen
from textual.widgets import Button, Footer, Header, Input, Markdown, Static, Switch

from nf_core.pipelines.create.utils import ShowLogs, TextInput, remove_hide_class

log = logging.getLogger(__name__)

github_org_help = """
> ⚠️  **You can't create a repository directly in the nf-core organisation.**
>
> Please create the pipeline repo to an organisation where you have access or use your user account.
> A core-team member will be able to transfer the repo to nf-core once the development has started.

> 💡 Your GitHub user account will be used by default if `nf-core` is given as the org name.
"""


class GithubRepo(Screen):
    """Create a GitHub repository and push all branches."""

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        gh_user, gh_token = self._get_github_credentials()
        github_text_markdown = dedent(
            """
            # Create GitHub repository

            Now that we have created a new pipeline locally, we can
            create a new GitHub repository and push the code to it.
            """
        )
        if gh_user:
            github_text_markdown += f">\n> 💡 _Found GitHub username {'and token ' if gh_token else ''}in local [GitHub CLI](https://cli.github.com/) config_\n>\n"
        yield Markdown(github_text_markdown)
        with Horizontal(classes="ghrepo-cols"):
            yield TextInput(
                "gh_username",
                "GitHub username",
                "Your GitHub username",
                default=gh_user[0] if gh_user is not None else "GitHub username",
                classes="column",
            )
            yield TextInput(
                "token",
                "GitHub token",
                "Your GitHub [link=https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/managing-your-personal-access-tokens]personal access token[/link] for login.",
                default=gh_token if gh_token is not None else "GitHub token",
                password=True,
                classes="column",
            )
            yield Button("Show", id="show_password")
            yield Button("Hide", id="hide_password")
        with Horizontal(classes="ghrepo-cols"):
            yield TextInput(
                "repo_org",
                "Organisation name",
                "The name of the organisation where the GitHub repo will be created",
                default=self.parent.TEMPLATE_CONFIG.org,
                classes="column",
            )
            yield TextInput(
                "repo_name",
                "Repository name",
                "The name of the new GitHub repository",
                default=self.parent.TEMPLATE_CONFIG.name,
                classes="column",
            )
        if self.parent.TEMPLATE_CONFIG.is_nfcore:
            yield Markdown(dedent(github_org_help))
        with Horizontal(classes="ghrepo-cols"):
            yield Switch(value=False, id="private")
            with Vertical():
                yield Static("Private", classes="")
                yield Static("Select to make the new GitHub repo private.", classes="feature_subtitle")
        yield Center(
            Button("Back", id="back", variant="default"),
            Button("Create GitHub repo", id="create_github", variant="success"),
            Button("Finish without creating a repo", id="exit", variant="primary"),
            classes="cta",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Create a GitHub repo or show help message and exit"""
        if event.button.id == "show_password":
            self.add_class("displayed")
            text_input = self.query_one("#token", TextInput)
            text_input.query_one(Input).password = False
        elif event.button.id == "hide_password":
            self.remove_class("displayed")
            text_input = self.query_one("#token", TextInput)
            text_input.query_one(Input).password = True
        elif event.button.id == "create_github":
            # Create a GitHub repo

            # Save GitHub username, token and repo name
            github_variables = {}
            for text_input in self.query("TextInput"):
                this_input = text_input.query_one(Input)
                github_variables[text_input.field_id] = this_input.value
            # Save GitHub repo config
            for switch_input in self.query("Switch"):
                github_variables[switch_input.id] = switch_input.value

            # Pipeline git repo
            pipeline_repo = git.Repo.init(
                Path(self.parent.TEMPLATE_CONFIG.outdir)
                / Path(self.parent.TEMPLATE_CONFIG.org + "-" + self.parent.TEMPLATE_CONFIG.name)
            )

            # GitHub authentication
            if github_variables["token"]:
                github_auth = self._github_authentication(github_variables["gh_username"], github_variables["token"])
            else:
                raise UserWarning(
                    f"Could not authenticate to GitHub with user name '{github_variables['gh_username']}'."
                    "Please provide an authentication token or set the environment variable 'GITHUB_AUTH_TOKEN'."
                )

            user = github_auth.get_user()
            org = None
            # Make sure that the authentication was successful
            try:
                user.login
                log.debug("GitHub authentication successful")
            except GithubException:
                raise UserWarning(
                    f"Could not authenticate to GitHub with user name '{github_variables['gh_username']}'."
                    "Please make sure that the provided user name and token are correct."
                )

            # Check if organisation exists
            # If the organisation is nf-core or it doesn't exist, the repo will be created in the user account
            if github_variables["repo_org"] != "nf-core":
                try:
                    org = github_auth.get_organization(github_variables["repo_org"])
                    log.info(
                        f"Repo will be created in the GitHub organisation account '{github_variables['repo_org']}'"
                    )
                except UnknownObjectException:
                    log.warn(f"Provided organisation '{github_variables['repo_org']}' not found. ")

            # Create the repo
            try:
                if org:
                    self._create_repo_and_push(
                        org,
                        github_variables["repo_name"],
                        pipeline_repo,
                        github_variables["private"],
                    )
                else:
                    # Create the repo in the user's account
                    log.info(
                        f"Repo will be created in the GitHub organisation account '{github_variables['gh_username']}'"
                    )
                    self._create_repo_and_push(
                        user,
                        github_variables["repo_name"],
                        pipeline_repo,
                        github_variables["private"],
                    )
            except UserWarning as e:
                log.error(f"There was an error with message: {e}")
                self.parent.push_screen("github_exit")

            self.parent.LOGGING_STATE = "repo created"
            self.parent.push_screen("logging")

    class RepoExists(Message):
        """Custom message to indicate that the GitHub repo already exists."""

        pass

    @on(RepoExists)
    def show_github_info_button(self) -> None:
        remove_hide_class(self.parent, "exit")
        remove_hide_class(self.parent, "back")

    @work(thread=True, exclusive=True)
    def _create_repo_and_push(self, org, repo_name, pipeline_repo, private):
        """Create a GitHub repository and push all branches."""
        self.post_message(ShowLogs())
        # Check if repo already exists
        try:
            repo = org.get_repo(repo_name)
            # Check if it has a commit history
            try:
                repo.get_commits().totalCount
                raise UserWarning(f"GitHub repository '{repo_name}' already exists")
            except GithubException:
                # Repo is empty
                repo_exists = True
            except UserWarning as e:
                # Repo already exists
                log.error(e)
                self.post_message(self.RepoExists())
                return
        except UnknownObjectException:
            # Repo doesn't exist
            repo_exists = False

        # Create the repo
        if not repo_exists:
            repo = org.create_repo(repo_name, description=self.parent.TEMPLATE_CONFIG.description, private=private)
            log.info(f"GitHub repository '{repo_name}' created successfully")
            remove_hide_class(self.parent, "close_app")

        # Add the remote
        try:
            pipeline_repo.create_remote("origin", repo.clone_url)
        except git.exc.GitCommandError:
            # Remote already exists
            pass
        # Push all branches
        pipeline_repo.remotes.origin.push(all=True).raise_if_error()

    def _github_authentication(self, gh_username, gh_token):
        """Authenticate to GitHub"""
        log.debug(f"Authenticating GitHub as {gh_username}")
        github_auth = Github(gh_username, gh_token)
        return github_auth

    def _get_github_credentials(self):
        """Get GitHub credentials"""
        gh_user = None
        gh_token = None
        # Use gh CLI config if installed
        gh_cli_config_fn = os.path.expanduser("~/.config/gh/hosts.yml")
        if os.path.exists(gh_cli_config_fn):
            try:
                with open(gh_cli_config_fn) as fh:
                    gh_cli_config = yaml.safe_load(fh)
                    gh_user = (gh_cli_config["github.com"]["user"],)
                    gh_token = gh_cli_config["github.com"]["oauth_token"]
            except KeyError:
                pass
        # If gh CLI not installed, try to get credentials from environment variables
        elif os.environ.get("GITHUB_TOKEN") is not None:
            gh_token = self.auth = os.environ["GITHUB_TOKEN"]
        return (gh_user, gh_token)
